module type OrderedType =
sig
    type t
    val compare: t -> t -> int
end

module Make (Vertex : OrderedType) = struct

  type vertex = Vertex.t
  type edges = vertex * vertex

  type graph = 
      { vertex : vertex array;
	vertex_hash : (vertex, int) Hashtbl.t;
	edges_hash : (edges, int) Hashtbl.t;
	adj : int array array;
	size : int }

  let size graph =
    graph.size 

  let add_edge graph vf vt w = 
    graph.adj.(Hashtbl.find graph.vertex_hash vf).(Hashtbl.find graph.vertex_hash vt) <- w;
    Hashtbl.add graph.edges_hash (vf, vt) w

  let weigth graph edge =
    try
      Hashtbl.find graph.edges_hash edge
    with Not_found ->
      -1
      
  let adjacent_list graph vertice =
    Array.fold_right (fun el acc -> 
			if weigth graph (vertice, el) >= 0 then
			  el :: acc
			else
			  acc) graph.vertex []

  let create vertex matrix =
    let vsize = Array.length vertex in
    let msize = Array.length matrix in
    let _ = if vsize <> msize then Debug.debug "Graphs.create bug" in 
    let vertex_hash = Hashtbl.create 0 in
    let _ = 
      for i = 0 to vsize - 1 do
	Hashtbl.add vertex_hash vertex.(i) i
      done in
    let edges = Hashtbl.create 0 in
      for i = 0 to vsize - 1 do
	for j = 0 to vsize - 1 do 
	  match matrix.(i).(j) with
	    | 0 -> ()
	    | n -> 
		Hashtbl.add edges (vertex.(i), vertex.(j)) n
	done
      done;
      { adj = matrix;
	edges_hash = edges;
	vertex = vertex;
	vertex_hash = vertex_hash;
	size = vsize;
      }


  let edges_of_list graph =
    Hashtbl.fold (fun key el acc -> (key, el) :: acc) graph.edges_hash []

  let sorted_edges graph =
    List.sort (fun (_, w1) (_, w2) -> compare w1 w2) (edges_of_list graph)

(*   let acm_kruskal graph = *)
(*     let my_size = size graph in *)
(*     let set = new Disjoints.t in *)
(*     let span = create graph.vertix (Array.create_matrix size 0) in *)
(*     let _ = iter_on_vertix (fun x -> set#add x) graph in *)
(*     let _ = List.iter  *)
(* 	      (fun ((vf, vt) as edge, _) -> *)
(* 		      if set#find vf = set#find cf then *)
(* 			begin *)
(* 			  add_edge span vf vt (weigth edge); *)
(* 			  set#union vf vt *)
(* 			end) in *)
(*       span *)

  let cost graph path = 
    let s = Array.length path in
    let c = ref 0 in
      for i = 0 to s - 2 do
	c := !c + weigth graph (path.(i),path.(i + 1))
      done;
      !c

  module MyPrioqueue = Prioqueue.Make (Vertex)
    

(* opt *)
  let acm_prim graph vertice =
    let f = new MyPrioqueue.t in
    let my_size = size graph in
    let _ = Array.iteri (fun n el ->
			   match n, el with
			       0, el -> f#add el (Prioqueue.Int 0)
			     | n, el -> f#add el Prioqueue.Inf)
	      graph.vertex in
    let father = Hashtbl.create 0 in
      while not f#empty do
	let u = f#extract_min in
	  List.iter (fun v -> 
		       let my_weight =  weigth graph (u,v) in
			 if f#belong v &&
			   Prioqueue.lesser (Prioqueue.Int my_weight) (f#priority v)
			 then begin
			   Hashtbl.replace father v u;
			   f#set_priority v (Prioqueue.Int my_weight)
			 end)
	    (adjacent_list graph u)
      done;
      father

  type 'a tree = Node of 'a * 'a tree list

(* opt *)

  let rec hash_to_tree hash racine =
    let suns vertice =
      Hashtbl.fold (fun key el acc ->
		      if el = vertice then
			key :: acc
		      else
			acc) hash [] in
    Node (racine, List.map (hash_to_tree hash) (suns racine))

      (* opt *)
  let salesman graph =
    if size graph = 0 then
      [||]
    else
      let first = graph.vertex.(0) in
      let spam_hash  = acm_prim graph first in
      let spam_tree = hash_to_tree spam_hash first in
      let l = ref [] in
      let rec parcours = function
	| Node (el, sons) ->
	    l := !l @ [el];
	    List.iter parcours sons in
	parcours spam_tree;
	(Array.of_list !l)

  let salesman_ex graph =
    let rec distribute elt = function
      | (hd :: tl) as list ->
	  (elt :: list) :: (List.map (fun x -> hd :: x) (distribute elt tl))
      | [] -> [[elt]] in
    let flat_rev_map f list =
      let rec loop acc = function
	| x :: rest -> loop (List.rev_append (f x) acc) rest
	| [] -> acc
      in
	loop [] list in
    let rec permute = function
      | x :: rest -> flat_rev_map (distribute x) (permute rest)
      | [] -> [[]] in
    match Array.to_list graph.vertex with
	[] -> [||]
      | (start :: tl) as l ->
	  let all_paths = List.map Array.of_list
			    (List.filter 
			       (fun l2 -> List.hd l2 = start) 
			       (permute l)) in
	    fst (List.fold_right (fun path (acc_p, acc_c) as acc -> 
				    let c = cost graph path in
				      if c < acc_c then
					(path, c)
				      else
					acc) all_paths ([||], max_int))
end
