open World
open Packages
open Robots
open Strategie

let write_to_socket oc s =
  Debug.print_out s;
  output_string oc (s ^ "\n");
  flush oc

let read_line_from_socket ic =
  let s = input_line ic in
    Debug.print_in s;
    s

let scanf_from_socket ic format f =
  let s = read_line_from_socket ic in
  Scanf.sscanf s format f

let write_player oc =
  write_to_socket oc "Player"

let read_size ic =
  scanf_from_socket ic "%d %d" (fun width height -> (width, height))

let tile_of_char = function
  | '.' -> World.Plain
  | '~' -> World.Lethal
  | '#' -> World.Impassable
  | '@' -> World.Home
  | _ -> failwith "tile_of_char" 

let read_map_line ic width =
  let line = read_line_from_socket ic in
  let map = Array.make (width + 2) World.Plain in
    map.(0) <- World.Impassable;
    for i = 0 to width - 1  do
      map.(i + 1) <- tile_of_char (line.[i])
    done;
    map.(width + 1) <- World.Impassable;
    map
    
let read_board ic width height =
  let walls =  Array.make (width + 2) World.Impassable in 
    Array.init (height + 2)
      (function 0 -> walls 
		  | n when n = height + 1 -> walls
		  | n ->  read_map_line ic width)

let read_init_param ic =
  scanf_from_socket ic "%d %d %d" (fun ruid capacity money ->
  (unsafe_rid_of_int ruid, capacity, money))
    
let rec parse_spaces = parser
  | [< '' '; s >] -> parse_spaces s
  | [< >] -> ()

let int_of_digit c =
  int_of_char c - int_of_char ('0')

let parse_number stream = 
  let rec parse_number_aux acc = parser
    | [< ''0' .. '9' as c; s >] ->
	parse_number_aux (10 * acc + (int_of_digit c)) s
    | [< >] -> acc in
    match stream with parser
	[< ''0' .. '9' as c; s >] -> parse_number_aux (int_of_digit c) s

let rec parse_list_of parse = parser
  | [< p = parse;
       _ = parse_spaces;
       ps = parse_list_of parse >] ->
      p :: ps
  | [< >] -> []

let parse_position = parser
    [<  
      ''X';
      _ = parse_spaces;
      x = parse_number;
      _ = parse_spaces;
      ''Y';
      _ = parse_spaces;
      y = parse_number >] -> (x, y)

let parse_init_position money = parser
    [<
      ''#';
      uid = parse_number;
      _ = parse_spaces;
      (x, y) = parse_position >] ->
      { money = money;
	ruid = unsafe_rid_of_int uid;
	rposition = (x, y);
	score = 0
      }

let parse_init_positions money =
  parse_list_of (parse_init_position money)

let read_init_positions ic money =
    (parse_init_positions money) (Stream.of_string (read_line_from_socket ic))

let parse_package ppos = parser
  [<
    id = parse_number;
    _ = parse_spaces;
    x = parse_number;
    _ = parse_spaces;
    y = parse_number;
    _ = parse_spaces;
    weight = parse_number >] ->
    { puid = id;
      weight = weight;
      destination = (x, y);
      plocation = Position ppos
    }

let parse_packages ppos = parse_list_of (parse_package ppos)

let read_packages w me ic ppos =
  List.filter (fun p ->
		 let comp = get_path_computer w w.start_pos in
		   is_reachable comp p.destination)
    ((parse_packages ppos) (Stream.of_string (read_line_from_socket ic)))

let init_world ic oc =
  let _ = write_player oc in
  let (width, height) = read_size ic in
  let board = read_board ic width height in
  let (ruid, capacity, money) = read_init_param ic in
  let init_pos = read_init_positions ic money in
  let robots = World.robots_of_list init_pos in
  let cst = {  map = board;
	       capacity = capacity;
	       initial_money = money;
	       width = width;
	       height = height;
	       start_pos =  World.robot_pos robots ruid;
	       paths = Weak.create (width * height);
	    } in
  let pack_list = read_packages cst ruid ic (World.robot_pos robots ruid) in
    ({  cst = cst;
       packages = Packages.packages_of_list pack_list;
       robots = robots;
       pushed_robots = r_clear_pushed;
       turn = 1;
     }, ruid)

let string_of_direction = function
  | N -> "N"
  | S -> "S"
  | E -> "E"
  | W -> "W"

let string_of_id_list l =
  String.concat " " (List.map string_of_int l)

let string_of_action = function
  | Move d ->
      "Move " ^ string_of_direction d
  | Pick l -> 
      "Pick " ^ string_of_id_list l
  | Drop l ->
      "Drop " ^ string_of_id_list l

let string_of_command (action, bid) =
  string_of_int bid ^ " " ^ string_of_action action

let write_command oc command = 
  write_to_socket oc (string_of_command command)

let update_packages ic world me =
  let pack_list = read_packages world.cst me ic (World.robot_pos
  world.robots me) in
    World.update_world_packages world pack_list

type server_action = Action of World.action | Appear of Utils.position

let parse_command = parser
  | [< ''E' >] -> Action (Move E)
  | [< ''W' >] -> Action (Move W)
  | [< ''N' >] -> Action (Move N)
  | [< ''S' >] -> Action (Move S)
  | [< ''P';
       _ = parse_spaces;
       n = parse_number >] -> Action (Pick [n])
  | [< ''D';
       _ = parse_spaces;
       n = parse_number >] -> Action (Drop [n])
  | [< pos = parse_position >] -> Appear (pos)

let parse_commands = parse_list_of parse_command

let parse_id_command = parser
  | [< ''#';
       id = parse_number;
       _ = parse_spaces;
       command = parse_commands >] ->
      (unsafe_rid_of_int id, command)

let parse_id_commands = parse_list_of parse_id_command

let read_commands ic =
  parse_id_commands (Stream.of_string (read_line_from_socket ic))

let update_world_with_command world (rid, command_list) =
  let aux world = function
  | Action (Move d) ->
      let cur_pos = (get_robot world rid).rposition in
      let new_pos = next_pos d cur_pos in
      let world =
	match robot_of_pos world new_pos with
	  | Some robot -> set_pushed world robot.ruid
	  | None -> world in
	World.move_robot world rid d
  | Action (Pick (pid :: [])) ->
	 World.change_world_package_loc world pid (Robot rid)
  | Action (Drop (pid :: [])) ->
      let r = get_robot world rid in
	(match get_package_by_id world.packages pid with
	   | Some p ->
	       if p.destination = r.rposition then
		 let world = replace_robot world { r with score = r.score + p.weight } in
		 let pkgs = rm_package_by_id world.packages p.puid in
		   { world with packages = pkgs }
	       else 
		 World.change_world_package_loc world pid 
		   (Position (World.robot_pos world.robots rid))
	   | None -> world)
  | Appear pos ->
      World.add_default_robot world rid pos
  | (Action (Pick _) | Action (Drop _)) ->
      failwith "update_world_with_command" in
    List.fold_left aux world command_list 
      
let update_world_with_commands ic world =
  let commands = read_commands ic in
    List.fold_left update_world_with_command world commands

let update_world ic oc world me command =
  write_command oc command;
  update_packages ic
    (update_world_with_commands ic
       (World.clear_pushed (World.update_turn world))) me
