open Packages

type tile =
  | Plain
  | Lethal
  | Impassable
  | Home

type worldcst = 
    { map : tile array array
    ; capacity : int
    ; initial_money : int
    ; width : int
    ; height : int
    ; start_pos : Utils.position
    ; paths :(Utils.position, int * Utils.position list) Hashtbl.t Weak.t
    }
  
(* world *)

type world = 
    { cst : worldcst
    ; packages : Packages.t
    ; robots : Robots.robot Robots.RobMap.t
    ; pushed_robots : Robots.IntSet.t
    ; turn : int
    }

let int_of_pos w (x, y) =
  (y - 1) * w.width + (x - 1)
  
let update_turn world =
  { world with turn = world.turn + 1 }

(* optimize *)
let nb_packages_robot w r =
  Packages.PacSet.cardinal (Packages.packages_on_robot w.packages r)

let change_world_package_loc world puid newloc =
  { world with packages = Packages.change_package_loc world.packages puid newloc }

let update_world_packages w l =
  { w with packages = Packages.add_packages w.packages l }

let all_homebases w = 
  let l = ref [] in
  for j = 1 to w.cst.height do
    for i = 1 to w.cst.width do
      if w.cst.map.(j).(i) = Home then l := (i, j) :: !l
    done
  done;
  !l

type direction = 
  | N 
  | S
  | E
  | W

type action = 
  | Move of direction
  | Pick of int list 
  | Drop of int list

type command =
    action * int

let get_tile w i j = w.map.(j).(i)
let set_tile w i j t = w.map.(j).(i) <- t

let next_pos dir (x, y) = 
  match dir with
    | N -> (x, y + 1)
    | S -> (x, y - 1)
    | W -> (x - 1, y)
    | E -> (x + 1, y)

(* optimize *)
let get_move (x1, y1) (x2, y2) =
  if x1 > x2 then W
  else if x1 < x2 then E
  else if y1 < y2 then N
  else S

(* robots *)

open Robots

let extract_me = Robots.r_extract_me

let add_robot w r = { w with robots =
			Robots.r_add_robot w.robots r }

let replace_robot w r = { w with robots =
			    Robots.r_replace_robot w.robots r }

let robots_of_list l =
  List.fold_left (fun acc r -> r_add_robot acc r) r_empty l

let new_default_robot w id pos =
  { money = w.cst.initial_money;
    ruid = id;
    rposition = pos;
    score = 0;
  }

let add_default_robot w id pos =
  replace_robot w (new_default_robot w id pos)

let get_robot w ruid = Robots.r_get_robot w.robots ruid

let robot_pos r id =
  (r_get_robot r id).rposition

let rm_robot w r =
  let others = Robots.r_rm_robot w.robots r in
    { w with robots = others;
	packages = Packages.fold_packages_on_robot w.packages r Packages.rm_package w.packages 
    }

let move_robot w id dir =
  let robot = get_robot w id in
  let pos = next_pos dir robot.rposition in
    {w with robots = Robots.r_add_robot w.robots 
		       { robot with rposition = pos }}

let set_pushed w id =
  { w with pushed_robots = Robots.r_set_pushed w.pushed_robots id }

let clear_pushed w =
  { w with pushed_robots = Robots.r_clear_pushed }

let was_pushed w id = Robots.r_was_pushed w.pushed_robots id

let robot_of_pos w pos =
  let l = Robots.r_fold_on_robots w.robots
	    (fun _ robot acc ->
	       if robot.rposition = pos then 
		 [robot]
	       else acc) [] in
    match l with
      |	hd :: [] -> Some hd
      | [] -> None
      | _ -> failwith "robot_of_pos"  

let money_of_robot w id =
  (Robots.r_get_robot w.robots id).money

(* optimize *)
let maybe_robot w (x, y) =
  Robots.r_fold_on_robots w.robots 
    (fun _ r acc -> 
       match acc with
	 | None -> if r.rposition = (x, y) then Some r else None 
	 | _ -> acc) None

let weight_of_robot w r =
  fold_packages_on_robot w.packages r (fun acc p -> acc + p.weight) 0 

let number_of_robot w =
  Robots.r_fold_on_robots w.robots
    (fun _ _ acc -> acc + 1) 0
